from .base import Crawler
from .chdbits import CHDBits
from .mteam import MTeam
from .fsm import FSM
from .ttg import TTG
from .pterclub import PTerClub
from .redleaves import RedLeaves
from .opencd import OpenCD
from .ourbits import OurBits
from .u2 import U2
from .ubits import UBits

VERSION = '1.0.6'

__all__ = ['VERSION']



from re import match
from math import inf
from typing import List
from typing import Dict
from typing import Optional
from http import HTTPStatus
from logging import Logger

from lxml import etree # pylint: disable=c-extension-no-member

from pydantic import ValidationError

from .base import Crawler
from .base import User
from .base import Torrent
from .base import Task
from .base import Promotion
from .base import get_id_from_href
from .base import find_element
from .base import Status
from .base import calculate_bytes
from .base import convert_to_bytes
from .exceptions import CannotGetUserInformationException
from .exceptions import CannotGetTorrentInformationException
from .exceptions import RequestException

promotion_map = {
    'free': Promotion(upload_ratio=1, download_ratio=0),
    'twoup': Promotion(upload_ratio=2, download_ratio=1),
    'twoupfree': Promotion(upload_ratio=2, download_ratio=0),
    'halfdown': Promotion(upload_ratio=1, download_ratio=0.5),
    'twouphalfdown': Promotion(upload_ratio=2, download_ratio=0.5),
    'thirtypercent': Promotion(upload_ratio=1, download_ratio=0.3),
}

def get_promotion_from_list(element: Optional[etree._Element]) -> Promotion: # pylint: disable=c-extension-no-member
    clazz = element.get('class') if element is not None else 'normal'
    if clazz == 'pro_50pctdown2up':
        return Promotion(upload_ratio=2, download_ratio=0.5)
    if clazz == 'pro_free':
        return Promotion(upload_ratio=1, download_ratio=0)
    if clazz == 'pro_free2up':
        return Promotion(upload_ratio=2, download_ratio=0)
    if clazz == 'pro_50pctdown':
        return Promotion(upload_ratio=1, download_ratio=0.5)
    if clazz == 'pro_30pctdown':
        return Promotion(upload_ratio=1, download_ratio=0.3)
    if clazz == 'pro_30pctdown':
        return Promotion(upload_ratio=1, download_ratio=0.3)
    return promotion_map.get(clazz, Promotion(upload_ratio=1, download_ratio=1))

def get_promotion_from_detail(element: Optional[etree._Element]) -> Promotion: # pylint: disable=c-extension-no-member
    clazz = element.get('class') if element is not None else 'normal'
    return promotion_map.get(clazz, Promotion(upload_ratio=1, download_ratio=1))

class RedLeaves(Crawler):
    def __init__(
        self,
        headers: Dict[str, str],
        base_url: str = 'https://leaves.red',
        proxy: Optional[str] = None,
        logger: Optional[Logger] = None,
        qps: float = inf,
        hr_policy: Optional[Dict[str, int]] = None,
    ) -> None:
        super().__init__(
            headers=headers,
            base_url=base_url,
            proxy=proxy,
            logger=logger,
            qps=qps
        )

    def get_user(self) -> User:
        pattern = r'.*'.join(
            [
                r'欢迎回来, (?P<user_name>.+)\ue901',
                rf'上传量: (?P<upload_number>{self.number_pattern}) (?P<upload_unit>{self.unit_pattern}) ',
                rf'下载量: (?P<download_number>{self.number_pattern}) (?P<download_unit>{self.unit_pattern}) ',
                rf'魔力值 : (?P<bonus>{self.number_pattern})'
            ]
        )

        response = self.session.get(url=self.base_url + '/usercp.php')
        if not response.status_code == HTTPStatus.OK:
            raise RequestException(response)

        html = etree.HTML(response.text) # pylint: disable=c-extension-no-member

        email_element = find_element(html, '//*[@id="outer"]/table[2]/tr[2]/td[2]')
        user_id_element = find_element(html, '//*[@id="info_block"]/tr/td/table/tr/td[1]/span/span[1]/a')
        title_element = find_element(html, '//*[@id="info_block"]/tr/td/table/tr/td[1]')
        passkey_element = find_element(html, '//*[@id="outer"]/table[2]/tr[4]/td[2]')

        if email_element is None or user_id_element is None or title_element is None or passkey_element is None:
            raise CannotGetUserInformationException()

        result = match(pattern, ' '.join(''.join(title_element.itertext()).split()))
        if not result:
            raise CannotGetUserInformationException()

        user_id = get_id_from_href(user_id_element.get('href'))
        if not user_id:
            raise CannotGetUserInformationException()

        return User(
            user_id=user_id,
            user_name=result.group('user_name'),
            upload_bytes=calculate_bytes(result.group('upload_number'), result.group('upload_unit')),
            download_bytes=calculate_bytes(result.group('download_number'), result.group('download_unit')),
            email=email_element.text,
            bonus=float(result.group('bonus').replace(',', '')),
            passkey=passkey_element.text
        )

    def get_torrents(self, pages: int = 1) -> List[Torrent]:
        torrents = []
        for page in range(pages):
            response = self.session.get(
                url=self.base_url + '/torrents.php',
                params={'page': str(page), 'incldead': '0', 'spstate': '0'}
            )

            if not response.status_code == HTTPStatus.OK:
                raise RequestException(response)

            html = etree.HTML(response.text) # pylint: disable=c-extension-no-member
            _, __, *rows = html.xpath('/html/body/table[2]/tr[2]/td/table/tr/td/table/tr')

            for row in rows:
                title_element = find_element(row, 'td[2]//a[1]')
                size_element = find_element(row, 'td[5]')
                seeders_element = find_element(row, 'td[6]')
                leechers_element = find_element(row, 'td[7]')
                promotion_element = find_element(row, './/img[starts-with(@class, "pro_")]')

                if title_element is None or size_element is None or seeders_element is None or leechers_element is None:
                    self.logger.warning(CannotGetTorrentInformationException())
                    continue

                torrent_id = get_id_from_href(title_element.get('href'))
                size = convert_to_bytes(' '.join(size_element.itertext()))

                if not torrent_id or not size:
                    self.logger.warning(CannotGetTorrentInformationException())
                    continue

                try:
                    torrent = Torrent(
                        torrent_id=torrent_id,
                        torrent_name=''.join(title_element.itertext()).strip(),
                        size=size,
                        seeders=int(''.join(seeders_element.itertext()).replace(',', '')),
                        leechers=int(''.join(leechers_element.itertext()).replace(',', '')),
                        hit_and_run=False,
                        promotion=get_promotion_from_list(promotion_element),
                        crawler=self,
                    )
                except ValidationError as exception:
                    self.logger.warning(exception)
                else:
                    torrents.append(torrent)

        return torrents

    def get_torrent(self, torrent_id: str) -> Torrent:
        response = self.session.get(
            url=self.base_url + '/details.php',
            params={'hit': '1', 'id': torrent_id}
        )

        if not response.status_code == HTTPStatus.OK:
            raise RequestException(response)

        html = etree.HTML(response.text) # pylint: disable=c-extension-no-member

        title_element = find_element(html, '//*[@id="top"]')
        base_information_element = find_element(html, "//td[preceding-sibling::td[1][text()='基本信息']]")
        seeder_and_leecher_element = find_element(html, "//td[preceding-sibling::td[1][text()='同伴']]")
        promotion_element = find_element(title_element, './/font[@class="free" or @class="twoup" or @class="twoupfree" or @class="halfdown" or @class="twouphalfdown" or @class="thirtypercent"]')

        if title_element is None or base_information_element is None or seeder_and_leecher_element is None:
            raise CannotGetTorrentInformationException()

        result = match(r'大小：(?P<size_number>.+)类型', ''.join(base_information_element.itertext()))
        if not result:
            raise CannotGetTorrentInformationException()
        size = convert_to_bytes(result.group('size_number').strip())

        torrent_name = title_element.text.strip()

        result = match(r'(?P<seeder_number>.+)个做种者 \| (?P<leecher_number>.+)个下载者', ''.join(seeder_and_leecher_element.itertext()))

        if not result or not size:
            raise CannotGetTorrentInformationException()

        return Torrent(
            torrent_id=torrent_id,
            torrent_name=torrent_name,
            size=size,
            seeders=int(result.group('seeder_number')),
            leechers=int(result.group('leecher_number')),
            promotion=get_promotion_from_detail(promotion_element),
            crawler=self,
            hit_and_run=False
        )

    def download_torrent(self, torrent_id: str, file_path: str) -> bool:
        response = self.session.get(
            url=self.base_url + '/download.php',
            params={'id': torrent_id}
        )

        if not response.status_code == HTTPStatus.OK:
            self.logger.warning(RequestException(response))
            return False

        with open(file_path, 'wb') as file:
            file.write(response.content)
            return True

    def get_tasks(self) -> List[Task]:
        user = self.get_user()
        tasks = []

        for status in [Status.LEECHING, Status.SEEDING]:
            response = self.session.get( # pylint: disable = missing-timeout
                url=self.base_url + '/getusertorrentlistajax.php',
                params={'userid': user.user_id, 'type': status},
            )

            if not response.status_code == HTTPStatus.OK:
                self.logger.warning(RequestException(response))
                continue

            html = etree.HTML(response.text) # pylint: disable = c-extension-no-member

            try:
                _, *rows = html.xpath('//tr')
            except ValueError:
                continue

            for row in rows:
                title_element = find_element(row, 'td[2]/a')
                href = title_element.get('href') or '' if title_element is not None else ''
                torrent_id = get_id_from_href(href)
                torrent_name = title_element.get('title') if title_element is not None else ''

                if not torrent_name or not torrent_id:
                    self.logger.warning('2333')
                    continue

                tasks.append(Task(torrent_id=torrent_id, torrent_name=torrent_name, status=status))

        return tasks



from typing import List
from typing import Dict
from typing import Optional
from http import HTTPStatus
from math import inf
from logging import Logger

from pydantic import BaseModel
from pydantic import Field
from pydantic import validator
from pydantic import ValidationError

from .base import Crawler
from .base import User
from .base import Task
from .base import Torrent
from .base import Status
from .base import Promotion
from .exceptions import RequestException

class SearchItemStatus(BaseModel):
    seeders: int
    leechers: int
    promotion: Promotion = Field(alias='discount')

    @validator('promotion', pre=True)
    def promotion_validator(cls, value: str) -> Promotion:
        if value == 'FREE':
            return Promotion(upload_ratio=1, download_ratio=0)

        if value == 'PERCENT_50':
            return Promotion(upload_ratio=1, download_ratio=0.5)

        if value == 'PERCENT_70':
            return Promotion(upload_ratio=1, download_ratio=0.3)

        if value == 'NORMAL':
            return Promotion(upload_ratio=1, download_ratio=1)

        return Promotion(upload_ratio=1, download_ratio=1)

class TaskTorrent(BaseModel):
    id: str
    name: str

class TaskItem(BaseModel):
    torrent: TaskTorrent

class TaskData(BaseModel):
    items: List[TaskItem] = Field(alias='data')
    page_number: int = Field(alias='pageNumber')
    total_pages: int = Field(alias='totalPages')

class TaskResponse(BaseModel):
    code: str
    message: str
    data: TaskData

class SearchItem(BaseModel):
    name: str
    size: int
    id: str
    status: SearchItemStatus

class SearchData(BaseModel):
    items: List[SearchItem] = Field(alias='data')

class SearchResponse(BaseModel):
    code: str
    message: str
    search_data: Optional[SearchData] = Field(alias='data')

class MemberCount(BaseModel):
    upload_bytes: int = Field(alias='uploaded')
    download_bytes: int = Field(alias='downloaded')
    bonus: float

class ProfileData(BaseModel):
    user_id: str = Field(alias='id')
    user_name: str = Field(alias='username')
    email: str
    member_count: MemberCount = Field(alias='memberCount')

class ProfileResponse(BaseModel):
    code: str
    message: str
    data: ProfileData

class TorrentData(BaseModel):
    id: str
    name: str
    size: int
    status: SearchItemStatus

class TorrentResponse(BaseModel):
    code: str
    message: str
    data: TorrentData

class DITokenResponse(BaseModel):
    code: str
    message: str
    download_url: str = Field(alias='data')

class MTeam(Crawler):
    def __init__(
        self,
        headers: Dict[str, str],
        base_url: str = 'https://api.m-team.cc',
        proxy: Optional[str] = None,
        logger: Optional[Logger] = None,
        qps: float = inf,
        hr_policy: Optional[Dict[str, int]] = None,
    ) -> None:
        super().__init__(
            headers=headers,
            base_url=base_url,
            proxy=proxy,
            logger=logger,
            qps=qps
        )

    def get_torrents(self, pages: int = 1, mode: str = 'normal') -> List[Torrent]:
        torrents = []

        for page in range(pages):
            response = self.session.post(
                url=self.base_url + '/api/torrent/search',
                json={
                    'mode': mode,
                    'categories': [],
                    'visible': 1,
                    'pageNumber': page + 1,
                    'pageSize': 100,
                }
            )

            if not response.status_code == HTTPStatus.OK:
                raise RequestException(response)

            search_response: SearchResponse = SearchResponse.parse_raw(response.text)
            if not search_response.search_data:
                self.logger.warning(search_response.message)
                break

            for item in search_response.search_data.items:
                try:
                    torrent = Torrent(
                        torrent_id=item.id,
                        torrent_name=item.name,
                        size=item.size,
                        hit_and_run=False,
                        promotion=item.status.promotion,
                        seeders=item.status.seeders,
                        leechers=item.status.leechers,
                        crawler=self,
                    )
                except ValidationError as exception:
                    self.logger.warning(exception)
                else:
                    torrents.append(torrent)

        return torrents

    def get_user(self) -> User:
        response = self.session.post(
            url=self.base_url + '/api/member/profile',
        )

        if not response.status_code == HTTPStatus.OK:
            raise RequestException(response)

        profile_response: ProfileResponse = ProfileResponse.parse_raw(response.text)
        return User(
            user_id=profile_response.data.user_id,
            user_name=profile_response.data.user_name,
            email=profile_response.data.email,
            upload_bytes=profile_response.data.member_count.upload_bytes,
            download_bytes=profile_response.data.member_count.download_bytes,
            bonus=profile_response.data.member_count.bonus,
        )

    def get_torrent(self, torrent_id: str) -> Torrent:
        response = self.session.post(self.base_url + '/api/torrent/detail', data={'id': torrent_id})

        if not response.status_code == HTTPStatus.OK:
            raise RequestException(response)

        torrent_response: TorrentResponse = TorrentResponse.parse_raw(response.text)
        return Torrent(
            torrent_id=torrent_response.data.id,
            torrent_name=torrent_response.data.name,
            size=torrent_response.data.size,
            hit_and_run=False,
            promotion=torrent_response.data.status.promotion,
            seeders=torrent_response.data.status.seeders,
            leechers=torrent_response.data.status.leechers,
            crawler=self
        )

    def download_torrent(self, torrent_id: str, file_path: str) -> bool:
        response = self.session.post(self.base_url + '/api/torrent/genDlToken', data={'id': torrent_id})

        if not response.status_code == HTTPStatus.OK:
            self.logger.warning(RequestException(response))
            return False

        di_token_response: DITokenResponse = DITokenResponse.parse_raw(response.text)
        response = self.session.get(di_token_response.download_url)

        if not response.status_code == HTTPStatus.OK:
            self.logger.warning(RequestException(response))
            return False

        with open(file_path, 'wb') as file:
            file.write(response.content)
            return True

    def get_tasks(self, page_size: int = 100) -> List[Task]:
        tasks = []
        user = self.get_user()

        for status in [Status.SEEDING, Status.LEECHING]:
            page_number = 1

            while True:
                response = self.session.post(
                    url=self.base_url + '/api/member/getUserTorrentList',
                    json={
                        "userid": user.user_id,
                        "type": status.upper(),
                        "pageNumber": page_number,
                        "pageSize": page_size
                    }
                )

                if not response.status_code == HTTPStatus.OK:
                    raise RequestException(response)

                task_response: TaskResponse = TaskResponse.parse_raw(response.text)
                for item in task_response.data.items:
                    task = Task(
                        torrent_id=item.torrent.id,
                        torrent_name=item.torrent.name,
                        status=status
                    )
                    tasks.append(task)

                if task_response.data.total_pages == page_number:
                    break

                page_number += 1

        return tasks



from math import inf
from time import time
from time import sleep

from requests import Session as BaseSession

class Session(BaseSession):
    def __init__(self, *args, qps: float = inf, **kwargs): # type: ignore
        super().__init__(*args, **kwargs)
        self.interval = 1 / qps
        self.last_request_time = time() - self.interval

    def request(self, *args,**kwargs): # type: ignore
        sleep_time = max(self.last_request_time + self.interval - time(), 0)
        sleep(sleep_time)

        response = super().request(*args, **kwargs)
        self.last_request_time = time()
        return response



from re import match
from math import inf
from typing import List
from typing import Dict
from typing import Optional
from http import HTTPStatus
from logging import Logger

from lxml import etree # pylint: disable=c-extension-no-member

from pydantic import ValidationError

from .base import Crawler
from .base import User
from .base import Torrent
from .base import Task
from .base import Promotion
from .base import get_id_from_href
from .base import find_element
from .base import Status
from .base import calculate_bytes
from .base import convert_to_bytes
from .exceptions import CannotGetUserInformationException
from .exceptions import CannotGetTorrentInformationException
from .exceptions import RequestException

def get_promotion(element: Optional[etree._Element]) -> Promotion: # pylint: disable=c-extension-no-member
    src = element.get('src') if element is not None else None

    if src == '/pic/ico_free.gif':
        return Promotion(upload_ratio=1, download_ratio=0)

    if src == '/pic/ico_30.gif':
        return Promotion(upload_ratio=1, download_ratio=0.3)

    if src == '/pic/ico_half.gif':
        return Promotion(upload_ratio=1, download_ratio=0.5)

    return Promotion(upload_ratio=1, download_ratio=1)
class TTG(Crawler):
    def __init__(
        self,
        headers: Dict[str, str],
        base_url: str = 'https://totheglory.im/',
        proxy: Optional[str] = None,
        logger: Optional[Logger] = None,
        qps: float = inf,
        hr_policy: Optional[Dict[str, int]] = None,
    ) -> None:
        super().__init__(
            headers=headers,
            base_url=base_url,
            proxy=proxy,
            logger=logger,
            qps=qps
        )

    def get_user(self) -> User:
        pattern = r'[\s\S]*'.join(
            [
                '',
                rf'上传量 :  (?P<upload_number>{self.number_pattern}) (?P<upload_unit>{self.unit_pattern})\xa0\xa0',
                rf'下载量 : (?P<download_number>{self.number_pattern}) (?P<download_unit>{self.unit_pattern})\xa0\xa0',
                rf'积分 : (?P<bonus>{self.number_pattern})\xa0\xa0'
            ]
        )

        response = self.session.get(
            url=self.base_url + '/my.php',
        )

        html = etree.HTML(response.content.decode('utf8')) # pylint: disable=c-extension-no-member

        email_element = find_element(html, '//*[@id="main_table"]/tr[1]/td/table/tr[2]/td/form/table/tr[5]/td[2]')
        passkey_element = find_element(html, '//*[@id="main_table"]/tr[1]/td/table/tr[2]/td/form/table/tr[6]/td[2]')
        title_element = find_element(html, '/html/body/table[2]/tr/td/table/tr/td[1]')
        user_id_element = find_element(html, '//*[@id="main_table"]/tr[1]/td/h1/a')

        if email_element is None or user_id_element is None or title_element is None or passkey_element is None:
            raise CannotGetUserInformationException()

        result = match(pattern, ''.join(title_element.itertext()))
        if not result:
            raise CannotGetUserInformationException()

        user_id = get_id_from_href(user_id_element.get('href'))
        if not user_id:
            raise CannotGetUserInformationException()

        return User(
            user_id=user_id,
            user_name=user_id_element.text,
            upload_bytes=calculate_bytes(result.group('upload_number'), result.group('upload_unit')),
            download_bytes=calculate_bytes(result.group('download_number'), result.group('download_unit')),
            email=email_element.text,
            bonus=float(result.group('bonus')),
            passkey=passkey_element.text
        )

    def get_torrents(self, pages: int = 1) -> List[Torrent]:
        torrents = []

        for page in range(pages):
            response = self.session.get(url=self.base_url + '/browse.php', params={'c': 'M', 'page': str(page)})
            if not response.status_code == HTTPStatus.OK:
                raise RequestException(response)

            html = etree.HTML(response.content.decode('utf8')) # pylint: disable=c-extension-no-member
            _, *rows = html.xpath('/html/body/table[3]/tr[1]/td/form/table/tr')
            for row in rows:
                torrent_id_element = find_element(row, './td[2]/div[1]/a')
                title_element = find_element(row, './td[2]/div[1]/a/b')
                size_element = find_element(row, './td[7]')
                seeders_and_leechers_element = find_element(row, './td[9]')

                if torrent_id_element is None or title_element is None or size_element is None or seeders_and_leechers_element is None:
                    self.logger.warning(CannotGetTorrentInformationException())
                    continue

                torrent_id_result = match(r'/t/(?P<torrent_id>.+)/', torrent_id_element.get('href') or '')
                size = convert_to_bytes(' '.join(size_element.itertext()))

                seeders_and_leechers_result = match(r'(?P<seeders>\d+)/\n(?P<leechers>\d+)', ''.join(seeders_and_leechers_element.itertext()))
                if not size or not torrent_id_result or not seeders_and_leechers_result or torrent_id_element is None or seeders_and_leechers_element is None:
                    self.logger.warning(CannotGetTorrentInformationException())
                    continue

                promotion_element = find_element(row, './/img[starts-with(@src, "/pic/ico_")]')
                hit_and_run_element = find_element(row, './/img[@title="Hit and Run"]')

                try:
                    torrent = Torrent(
                        torrent_id=torrent_id_result.group('torrent_id'),
                        torrent_name=title_element.text,
                        size=size,
                        seeders=seeders_and_leechers_result.group('seeders'),
                        leechers=seeders_and_leechers_result.group('leechers'),
                        hit_and_run=60 * 3600 if hit_and_run_element is not None else 0,
                        promotion=get_promotion(promotion_element),
                        crawler=self,
                    )
                except ValidationError as exception:
                    self.logger.warning(exception)
                    continue
                else:
                    torrents.append(torrent)

        return torrents

    def get_torrent(self, torrent_id: str) -> Torrent:
        response = self.session.get(self.base_url + f'/t/{torrent_id}/')
        if not response.status_code == HTTPStatus.OK:
            raise RequestException(response)

        html = etree.HTML(response.content.decode('utf8')) # pylint: disable=c-extension-no-member

        title_element = find_element(html, '//*[@id="main_table"]/tr[1]/td/h1')
        size_element = find_element(html, "//td[preceding-sibling::td[1][text()='尺寸']]")
        seeders_and_leechers_element = find_element(html, "//td[preceding-sibling::td[1][text()='活跃用户']]")

        if title_element is None or size_element is None or seeders_and_leechers_element is None:
            raise CannotGetTorrentInformationException()

        size_result = match(rf'(?P<size_number>{self.number_pattern}) (?P<size_unit>{self.unit_pattern})', size_element.text)
        seeders_and_leechers_result = match(r'(?P<seeders>.+) 做种者，(?P<leechers>.+) 下载者', ''.join(seeders_and_leechers_element.itertext()))
        if not size_result or not seeders_and_leechers_result:
            raise CannotGetTorrentInformationException()

        hit_and_run_element = find_element(html, './/img[@alt="Hit & Run"]')
        promotion_element = find_element(html, './/img[starts-with(@src, "/pic/ico_")]')

        return Torrent(
            torrent_id=torrent_id,
            torrent_name=title_element.text,
            size=calculate_bytes(size_result.group('size_number'), size_result.group('size_unit')),
            seeders=seeders_and_leechers_result.group('seeders'),
            leechers=seeders_and_leechers_result.group('leechers'),
            hit_and_run=60 * 3600 if hit_and_run_element is not None else 0,
            promotion=get_promotion(promotion_element),
            crawler=self,
        )

    def download_torrent(self, torrent_id: str, file_path: str) -> bool:
        response = self.session.get(self.base_url + f'/t/{torrent_id}/')
        if not response.status_code == HTTPStatus.OK:
            self.logger.warning(RequestException(response))
            return False

        html = etree.HTML(response.content.decode('utf8')) # pylint: disable=c-extension-no-member
        download_url_element = find_element(html, '//*[@id="main_table"]/tr[1]/td/table[1]/tr[1]/td[2]/a[1]')
        if download_url_element is None:
            self.logger.warning(CannotGetTorrentInformationException())
            return False

        href = download_url_element.get('href')
        if download_url_element is None:
            self.logger.warning(CannotGetTorrentInformationException())
            return False

        response = self.session.get(self.base_url + href)
        if not response.status_code == HTTPStatus.OK:
            self.logger.warning(RequestException(response))
            return False

        with open(file_path, 'wb') as file:
            file.write(response.content)
            return True

    def get_tasks(self) -> List[Task]:
        user = self.get_user()

        response = self.session.get(self.base_url + '/userdetails.php', params={'id': user.user_id})
        if not response.status_code == HTTPStatus.OK:
            raise RequestException(response)

        html = etree.HTML(response.content.decode('utf8')) # pylint: disable=c-extension-no-member

        tasks = []
        for keyword, status in [('当前上传', Status.SEEDING), ('当前下载', Status.LEECHING)]:
            try:
                _, *rows = html.xpath(f'''//td[preceding-sibling::td[1][text()={repr(keyword)}]]//tr''')
            except ValueError:
                continue

            for row in rows:
                title_element = find_element(row, './/td[2]/a/b')
                link_element = find_element(row, './/td[2]/a')

                if title_element is None or link_element is None:
                    self.logger.warning(CannotGetTorrentInformationException())
                    continue

                torrent_id = get_id_from_href(self.base_url + link_element.get('href', '/'))
                tasks.append(Task(torrent_id=torrent_id, torrent_name=title_element.text, status=status))
        return tasks



from requests import Response

class CannotGetUserInformationException(Exception):
    def __init__(self) -> None:
        super().__init__('cannot get user information')

class RequestException(Exception):
    def __init__(self, response: Response) -> None:
        super().__init__(f'request error: {response.status_code} {response.reason}')

class CannotGetTorrentInformationException(Exception):
    def __init__(self) -> None:
        super().__init__('cannot get torrent information')
